#include "http.h"
#include "web-server-base.h"
#include "web-server-this.h"
#include "web.h"
#include "web-pages-base.h"
#include "mstimer.h"

#define LOGIN_DELAY_MS 200

#define DO_LOGIN DO_SERVER + 0

struct state
{
    int      toDo;
    bool     postComplete;
    uint32_t delayUntil;
};

static int decideWhatToDo(char *pPath, char* pLastModified)
{
    if (HttpSameStr(pPath, "/login")) return DO_LOGIN;
    
    int todo;
    todo = WebServerBaseDecideWhatToDo(pPath, pLastModified); if (todo != DO_NOT_FOUND) return todo;
    todo = WebServerThisDecideWhatToDo(pPath, pLastModified); if (todo != DO_NOT_FOUND) return todo;
    return DO_NOT_FOUND;
}
static void handleQuery(int todo, char* pQuery)
{
    switch (todo)
    {
        case DO_LOGIN:         WebLoginQuery   (pQuery); return;
    }
    if (WebServerBaseHandleQuery(todo, pQuery)) return;
    if (WebServerThisHandleQuery(todo, pQuery)) return;
}
static void handlePost(int todo, int contentLength, int contentStart, int size, char* pRequestStream, uint32_t positionInRequestStream, bool* pComplete)
{
    if (WebServerBasePost(todo, contentLength, contentStart, size, pRequestStream, positionInRequestStream, pComplete)) return;
    if (WebServerThisPost(todo, contentLength, contentStart, size, pRequestStream, positionInRequestStream, pComplete)) return;
    *pComplete = true;
}

static void reply(int todo)
{
    //Try all the base modules
    switch (todo)
    {
        case DO_LOGIN:           WebLoginHtml     (); return;
        case DO_NOT_FOUND:       HttpNotFound     (); return;
        case DO_NOT_MODIFIED:    HttpNotModified  (); return;
    }
    
    //If not called then call the derived (child) module
    if (WebServerBaseReply(todo)) return;
    if (WebServerThisReply(todo)) return;
}

static void handleRequest(char* pStateData, int size, char* pRequestStream, uint32_t positionInRequestStream)
{
    struct state* pState = (struct state*)pStateData;
    
    pState->delayUntil = MsTimerCount; //Default to no delay unless modified;
    
    //Handle request for the first packet of data received but leave todo the same after that.
    int contentLength = 0;
    int contentStart  = 0;
    if (size && positionInRequestStream == 0)
    {
        //Read the headers
        char* pMethod;
        char* pPath;
        char* pQuery;
        char* pLastModified;
        char* pCookies;
        contentStart = HttpRequestRead(pRequestStream, size, &pMethod, &pPath, &pQuery, &pLastModified, &pCookies, &contentLength);
        
        //Ask the web server what to do
        pState->toDo = decideWhatToDo(pPath, pLastModified);
        
        //If what to do is NOTHING, NOT_FOUND or NOT_MODIFIED then no query or post will be valid so stop now
        if (pState->toDo < DO_LOGIN) { pState->postComplete = true; return; }
        
        //If what to do is LOGIN then the user has just returned the login form
        if (pState->toDo == DO_LOGIN)
        {
            handleQuery(pState->toDo, pQuery);                  //Read the password and the original location
            if (WebLoginQueryPasswordOk)
            {
                if (!WebLoginSessionIdIsSet())           //If there isn't a session id already
                {
                    WebLoginSessionIdNew();              //Create a new session id
                }
                pState->toDo =  WebLoginOriginalToDo;          //Load the original todo and SEND_SESSION_ID
                pState->toDo += DO_SEND_SESSION_ID;
            }
            pState->delayUntil = MsTimerCount + LOGIN_DELAY_MS; //To prevent brute forcing the hash delay the reply to the login
            pState->postComplete = true;
            return;                                       //Either way no query or post will be valid
        }
        
        //Have a normal request so authenticate
        if (!WebLoginCookiesContainValidSessionId(pCookies))
        {
            WebLoginOriginalToDo = pState->toDo; //Record the original destination for redirection
            pState->toDo = DO_LOGIN;
            pState->postComplete = true;
            return; //Ignore any query or post as the user is not authenticated
        }
        
        //Handle the query
        handleQuery(pState->toDo, pQuery);
    }
    
    //Do the upload of anything that needs it. Todos it doesn't understand are ignored.
    if (!pState->postComplete) handlePost(pState->toDo, contentLength, contentStart, size, pRequestStream, positionInRequestStream, &pState->postComplete);
}

static int pollSomethingToSend(char* pStateData, bool clientFinished) //returns 0 no; +1 yes; -1 finished; 
{
    struct state* pState = (struct state*)pStateData;
    
    if (!pState->toDo)
    {
        if (clientFinished) return -1; //The client hasn't requested anything and never will so finish
        else                return  0; //The client hasn't requested anything yet but still could
    }
    if (!pState->postComplete               ) return 0; //Wait for the request (usually a POST) to finish
    if (!MsTimerAbsolute(pState->delayUntil)) return 0; //Wait a while (usually after a LOGIN attempt)
    
    return 1; //Yep, set up anything (just TSL at the moment) then call sendReply
}

static bool sendReply(char* pStateData)
{
    struct state* pState = (struct state*)pStateData;
    
    int todo = pState->toDo; //Make a copy so that we don't modify todo in the state

    //Check if todo includes the need to send a cookie
    if (todo >= DO_SEND_SESSION_ID)
    {
        HttpOkCookieName   = WebLoginSessionNameGet();
        HttpOkCookieValue  = WebLoginSessionIdGet();
        HttpOkCookieMaxAge = WebLoginSessionNameLife();
        todo -= DO_SEND_SESSION_ID;
    }
    else
    {
        HttpOkCookieName   = NULL;
        HttpOkCookieValue  = NULL;
        HttpOkCookieMaxAge = -1;
    }
    
    reply(todo);
    
    return !HttpBufFilled(); //If we haven't used a full buffer then we have finished
}

int WebInit()
{
    HttpRequestFunction = handleRequest;
    HttpPollFunction    = pollSomethingToSend;
    HttpReplyFunction   = sendReply;
    
    WebLoginInit();
    
    return 0;
}